"""MGZ database API."""

import io
import logging
import os
import time
from datetime import timedelta, datetime

import pkg_resources
import requests
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from sqlalchemy.exc import IntegrityError

import mgz.summary

from mgzdb.platforms import PLATFORM_VOOBLY
from mgzdb.schema import (
    Match, SeriesMetadata, File, Player,
    Team, User, Series, Dataset, EventMap, Chat,
    Tournament, Round
)
from mgzdb.util import parse_filename, save_file, get_unique
from mgzdb.compress import compress, compress_tiles
from mgzdb.extract import save_starting_state, save_extraction
from mgz.util import Version
from mgz.summary.chat import Chat as ChatType


LOGGER = logging.getLogger(__name__)
LOG_ID_LENGTH = 8
COMPRESSED_EXT = '.mgc'


def has_transposition(user_data, players):
    """Check for player data transposition.

    Occasionally platforms will incorrectly record
    match metadata, so we check it against the rec
    to make sure it's accurate.
    """
    if not user_data:
        return False
    from_rec = {}
    for player in players:
        from_rec[player['color_id']] = player['name']
    from_user_data = {}
    for user in user_data:
        from_user_data[user['color_id']] = user['username']
    strike = 0
    for color_id, name in from_rec.items():
        if (
                color_id not in from_user_data or
                from_user_data[color_id].lower() not in name.lower() and
                name.lower() not in from_user_data[color_id].lower()
        ):
            strike += 1
    if strike >= len(players) / 2:
        return True
    return False


def merge_platform_attributes(ladder, platform_id, match_id, data, platforms):
    """Merge various platform attributes."""
    rated = data['rated']
    if ladder:
        rated = True
    if not ladder:
        ladder = data['ladder']
    if data['platform_id'] and not platform_id:
        platform_id = data['platform_id']
    if data['platform_match_id'] and not match_id:
        match_id = data['platform_match_id']
    ladder_id = None
    if platform_id:
        try:
            ladder_id = platforms[platform_id].lookup_ladder_id(ladder)
        except (KeyError, ValueError):
            ladder_id = None
    return rated, ladder_id, platform_id, match_id


def file_exists(session, file_hash, series_id):
    """Check if a file exists.

    Update the series_id if necessary.
    """
    exists = session.query(File).filter_by(hash=file_hash).one_or_none()
    if exists:
        if not exists.match.series_id and series_id:
            data = session.query(Tournament.event_id, Tournament.id).join(Round).join(Series).filter(Series.id == series_id).one_or_none()
            exists.match.event_id = data[0]
            exists.match.tournament_id = data[1]
            exists.match.series_id = series_id
            session.commit()
        return exists.match_id
    return False


def save_chat(session, chats, match_id):
    """Save chat messages."""
    objs = []
    for chat in chats:
        if chat['type'] != ChatType.MESSAGE:
            continue
        del chat['type']
        chat['timestamp'] = timedelta(milliseconds=chat['timestamp'])
        objs.append(Chat(match_id=match_id, **chat))
    session.bulk_save_objects(objs)
    session.commit()


class AddFile:
    """Add file to MGZ Database."""

    def __init__(self, session, platforms, store_path, playback):
        """Initialize sessions."""
        self.session = session
        self.platforms = platforms
        self.store_path = store_path
        self.playback = playback

    def add_file( # pylint: disable=too-many-return-statements, too-many-branches
            self, rec_path, reference, series_name=None,
            series_id=None, platform_id=None, platform_match_id=None,
            platform_metadata=None, played=None, ladder=None, user_data=None
    ):
        """Add a single mgz file."""
        start = time.time()
        if not os.path.isfile(rec_path):
            LOGGER.error("%s is not a file", rec_path)
            return False, 'Not a file'

        original_filename = os.path.basename(rec_path)
        modified = datetime.fromtimestamp(os.path.getmtime(rec_path))

        with open(rec_path, 'rb') as handle:
            data = handle.read()

        try:
            handle = io.BytesIO(data)
            summary = mgz.summary.Summary(handle, self.playback)
            # Hash against body only because header can vary based on compression
            file_hash = summary.get_file_hash()
            log_id = file_hash[:LOG_ID_LENGTH]
            LOGGER.info("[f:%s] add started", log_id)
        except RuntimeError as error:
            LOGGER.error("[f] invalid mgz file: %s", str(error))
            return False, 'Invalid mgz file'
        except LookupError:
            LOGGER.error("[f] unknown encoding")
            return False, 'Unknown text encoding'
        except ValueError as error:
            LOGGER.error("[f] error: %s", error)
            return False, error

        existing_match_id = file_exists(self.session, file_hash, series_id)
        if existing_match_id:
            LOGGER.warning("[f:%s] file already exists (%d)", log_id, existing_match_id)
            self._handle_file(file_hash, data, Version(summary.get_version()[0]))
            return None, existing_match_id

        try:
            encoding = summary.get_encoding()
        except ValueError as error:
            LOGGER.error("[f] error: %s", error)
            return False, error
        match_hash_obj = summary.get_hash()
        if not match_hash_obj:
            LOGGER.error("f:%s] not enough data to calculate safe match hash", log_id)
            return False, 'Not enough data to calculate safe match hash'
        match_hash = match_hash_obj.hexdigest()
        build = None

        try:
            if not platform_match_id and summary.get_platform()['platform_match_id']:
                platform_match_id = summary.get_platform()['platform_match_id']
            where = (Match.hash == match_hash)
            if platform_match_id:
                where |= (Match.platform_match_id == platform_match_id)
            match = self.session.query(Match).filter(where).one()
            LOGGER.info("[f:%s] match already exists (%d); appending", log_id, match.id)
        except MultipleResultsFound:
            LOGGER.error("[f:%s] mismatched hash and match id: %s, %s", log_id, match_hash, platform_match_id)
            return False, 'Mismatched hash and match id'
        except NoResultFound:
            LOGGER.info("[f:%s] adding match", log_id)
            parsed_played, build = parse_filename(original_filename)
            if not played:
                played = parsed_played
            try:
                match, message = self._add_match(
                    summary, played, match_hash, user_data, series_name,
                    series_id, platform_id, platform_match_id,
                    platform_metadata, ladder, build
                )
                if not match:
                    return False, message
                self._update_match_users(platform_id, match.id, user_data)
                self._update_match_de(match)
            except IntegrityError:
                LOGGER.error("[f:%s] constraint violation: could not add match", log_id)
                return False, 'Failed to add match'

        compressed_filename, compressed_size = self._handle_file(file_hash, data, Version(match.version_id))

        try:
            new_file = get_unique(
                self.session,
                File, ['hash'],
                filename=compressed_filename,
                original_filename=original_filename,
                hash=file_hash,
                size=summary.size,
                modified=modified,
                compressed_size=compressed_size,
                encoding=encoding,
                language=summary.get_language(),
                reference=reference,
                match=match,
                owner_number=summary.get_owner(),
                parser_version=pkg_resources.get_distribution('mgz').version
            )
            self.session.add(new_file)
            self.session.commit()
        except RuntimeError:
            LOGGER.error("[f:%s] unable to add file, likely hash collision", log_id)
            return False, 'File hash collision'

        LOGGER.info("[f:%s] add finished in %.2f seconds, file id: %d, match id: %d",
                    log_id, time.time() - start, new_file.id, match.id)
        return file_hash, match.id

    def _add_match( # pylint: disable=too-many-branches, too-many-return-statements
            self, summary, played, match_hash, user_data,
            series_name=None, series_id=None, platform_id=None,
            platform_match_id=None, platform_metadata=None, ladder=None, build=None
    ):
        """Add a match."""
        try:
            postgame = summary.get_postgame()
            duration = summary.get_duration()
        except RuntimeError:
            LOGGER.error("[m] failed to get duration")
            return False, 'Failed to get duration'

        log_id = match_hash[:LOG_ID_LENGTH]
        platform_data = summary.get_platform()
        rated, ladder_id, platform_id, platform_match_id = merge_platform_attributes(ladder, platform_id, platform_match_id, platform_data, self.platforms)

        settings = summary.get_settings()
        try:
            map_data = summary.get_map()
        except ValueError:
            LOGGER.error("[m:%s] has an invalid map", log_id)
            return False, 'Invalid map'
        completed = summary.get_completed()
        restored, _ = summary.get_restored()
        has_postgame = bool(postgame)
        version_id, game_version, save_version = summary.get_version()
        try:
            dataset_data = summary.get_dataset()
        except ValueError:
            LOGGER.error("[m:%s] dataset inconclusive", log_id)
            return False, 'Inconclusive dataset'

        teams = summary.get_teams()
        diplomacy = summary.get_diplomacy()
        players = list(summary.get_players())
        mirror = summary.get_mirror()
        if platform_match_id:
            log_id += ':{}'.format(platform_match_id)

        if restored:
            LOGGER.error("[m:%s] is restored game", log_id)
            return False, 'Restored matches not supported'

        if not completed:
            LOGGER.error("[m:%s] was not completed", log_id)
            return False, 'Incomplete matches not supported'

        if user_data and len(players) != len(user_data):
            LOGGER.error("[m:%s] has mismatched user data", log_id)
            return False, 'Mismatched user data'

        if has_transposition(user_data, players):
            LOGGER.error("[m:%s] has mismatched user data (transposition)", log_id)
            return False, 'Transposed user data'

        try:
            dataset = self.session.query(Dataset).filter_by(id=dataset_data['id']).one()
        except NoResultFound:
            LOGGER.error("[m:%s] dataset not supported: userpatch id: %s (%s)",
                         log_id, dataset_data['id'], dataset_data['name'])
            return False, 'Dataset not supported'

        series, tournament, event, event_map_id = self._handle_series(series_id, series_name, map_data, log_id)
        if series:
            played = series.played

        objects = summary.get_objects()

        match = get_unique(
            self.session,
            Match, ['hash', 'platform_match_id'],
            platform_match_id=platform_match_id,
            platform_id=platform_id,
            platform_metadata=platform_metadata,
            played=played,
            hash=match_hash,
            series=series,
            tournament=tournament,
            event=event,
            version_id=version_id.value,
            game_version=game_version,
            save_version=round(save_version, 2),
            build=build,
            dataset=dataset,
            dataset_version=dataset_data['version'],
            ladder_id=ladder_id,
            rated=rated,
            lobby_name=platform_data['lobby_name'],
            builtin_map_id=map_data['id'],
            event_map_id=event_map_id,
            map_size_id=map_data['dimension'],
            map_name=map_data['name'],
            map_tiles=compress_tiles(map_data['tiles']),
            rms_seed=map_data['seed'],
            rms_zr=map_data['zr'],
            rms_custom=map_data['custom'],
            guard_state=map_data['modes']['guard_state'],
            direct_placement=map_data['modes']['direct_placement'],
            effect_quantity=map_data['modes']['effect_quantity'],
            fixed_positions=map_data['modes']['fixed_positions'],
            water_percent=map_data['water'],
            duration=timedelta(milliseconds=duration),
            completed=completed,
            restored=restored,
            postgame=has_postgame,
            type_id=settings['type'][0],
            difficulty_id=settings['difficulty'][0],
            population_limit=settings['population_limit'],
            map_reveal_choice_id=settings['map_reveal_choice'][0],
            speed_id=settings['speed'][0],
            cheats=settings['cheats'],
            lock_teams=settings['lock_teams'],
            treaty_length=settings['treaty_length'],
            starting_resources_id=settings['starting_resources'][0],
            starting_age_id=settings['starting_age'][0],
            victory_condition_id=settings['victory_condition'][0],
            team_together=settings['team_together'],
            all_technologies=settings['all_technologies'],
            lock_speed=settings['lock_speed'],
            multiqueue=settings['multiqueue'],
            diplomacy_type=diplomacy['type'],
            team_size=diplomacy.get('team_size'),
            mirror=mirror,
            starting_town_centers=objects.get('tcs'),
            starting_walls=objects.get('stone_walls'),
            starting_palisades=objects.get('palisade_walls')
        )

        winning_team_id = None
        for data in players:
            team_id = None
            for i, team in enumerate(teams):
                if data['number'] in team:
                    team_id = i
            if data['winner']:
                winning_team_id = team_id
            feudal_time = data['achievements']['technology']['feudal_time']
            castle_time = data['achievements']['technology']['castle_time']
            imperial_time = data['achievements']['technology']['imperial_time']
            try:
                get_unique(
                    self.session,
                    Team,
                    ['match', 'team_id'],
                    winner=(team_id == winning_team_id),
                    match=match,
                    team_id=team_id
                )
                if data['user_id']:
                    get_unique(self.session, User, ['id', 'platform_id'], id=str(data['user_id']), platform_id=platform_id)
                player = get_unique(
                    self.session,
                    Player,
                    ['match_id', 'number'],
                    civilization_id=int(data['civilization']),
                    team_id=team_id,
                    match_id=match.id,
                    dataset=dataset,
                    platform_id=platform_id,
                    user_id=data['user_id'],
                    user_name=data['name'] if data['user_id'] else None,
                    human=data['human'],
                    name=data['name'],
                    number=data['number'],
                    color_id=data['color_id'],
                    start_x=data['position'][0],
                    start_y=data['position'][1],
                    winner=data['winner'],
                    mvp=data['mvp'],
                    cheater=data['cheater'],
                    score=data['score'],
                    rate_snapshot=data['rate_snapshot'],
                    military_score=data['achievements']['military']['score'],
                    units_killed=data['achievements']['military']['units_killed'],
                    hit_points_killed=data['achievements']['military']['hit_points_killed'],
                    units_lost=data['achievements']['military']['units_lost'],
                    buildings_razed=data['achievements']['military']['buildings_razed'],
                    hit_points_razed=data['achievements']['military']['hit_points_razed'],
                    buildings_lost=data['achievements']['military']['buildings_lost'],
                    units_converted=data['achievements']['military']['units_converted'],
                    economy_score=data['achievements']['economy']['score'],
                    food_collected=data['achievements']['economy']['food_collected'],
                    wood_collected=data['achievements']['economy']['wood_collected'],
                    stone_collected=data['achievements']['economy']['stone_collected'],
                    gold_collected=data['achievements']['economy']['gold_collected'],
                    tribute_sent=data['achievements']['economy']['tribute_sent'],
                    tribute_received=data['achievements']['economy']['tribute_received'],
                    trade_gold=data['achievements']['economy']['trade_gold'],
                    relic_gold=data['achievements']['economy']['relic_gold'],
                    technology_score=data['achievements']['technology']['score'],
                    feudal_time=timedelta(seconds=feudal_time) if feudal_time else None,
                    castle_time=timedelta(seconds=castle_time) if castle_time else None,
                    imperial_time=timedelta(seconds=imperial_time) if imperial_time else None,
                    explored_percent=data['achievements']['technology']['explored_percent'],
                    research_count=data['achievements']['technology']['research_count'],
                    research_percent=data['achievements']['technology']['research_percent'],
                    society_score=data['achievements']['society']['score'],
                    total_wonders=data['achievements']['society']['total_wonders'],
                    total_castles=data['achievements']['society']['total_castles'],
                    total_relics=data['achievements']['society']['total_relics'],
                    villager_high=data['achievements']['society']['villager_high']
                )
            except RuntimeError:
                LOGGER.warning("[m:%s] failed to insert players (probably bad civ id: %d)", log_id, data['civilization'])
                return False, 'Failed to load players'

            if match.platform_id == PLATFORM_VOOBLY and not user_data:
                self._guess_match_user(player, data['name'])


        match.winning_team_id = winning_team_id
        success, attrs = save_extraction(self.session, summary, ladder_id, match.id, dataset_data['id'], log_id)
        if success:
            match.has_playback = True
            match.state_reader_version = attrs['version']
            match.state_reader_interval = attrs['interval']
            match.state_reader_runtime = attrs['runtime']
        else:
            save_starting_state(self.session, objects['objects'], match.id, dataset_data['id'])

        save_chat(self.session, summary.get_chat(), match.id)
        return match, None

    def _update_match_de(self, match):
        if match.platform_id != 'de':
            return
        resp = requests.get('https://aoe2.net/api/match?uuid={}'.format(match.platform_match_id))
        if resp.status_code != 200:
            LOGGER.warning('failed to get aoe2.net data')
            return
        data = resp.json()
        match.played = datetime.fromtimestamp(data['started'])
        match.rated = data['ranked']
        match.lobby_private = data['has_password']
        match.lobby_opened = datetime.fromtimestamp(data['opened'])
        match.server = data['server']
        match.ladder_id = data['leaderboard_id']
        if data['version']:
            match.build = '101.101.{}.0'.format(data['version'])
        for p in data['players']:
            try:
                player = self.session.query(Player).filter_by(match_id=match.id, user_id=str(p['profile_id'])).one()
            except NoResultFound:
                LOGGER.error("failed to find p%d to update platform user data", p['color'])
                continue
            LOGGER.info("[m:%s] updating platform user data for p%d", match.hash[:LOG_ID_LENGTH], p['color'])
            get_unique(self.session, User, ['id', 'platform_id'], id=str(p['profile_id']), platform_id=match.platform_id)
            player.rate_snapshot = p['rating']


    def _update_match_users(self, platform_id, match_id, user_data):
        """Update Voobly User info on Match."""
        if not user_data:
            return
        for user in user_data:
            try:
                player = self.session.query(Player).filter_by(match_id=match_id, color_id=user['color_id']).one()
            except NoResultFound:
                LOGGER.error("failed to find p%d to update platform user data",
                             user['color_id'] + 1)
                continue
            LOGGER.info("[m:%s] updating platform user data for p%d",
                        player.match.hash[:LOG_ID_LENGTH], user['color_id'] + 1)
            get_unique(self.session, User, ['id', 'platform_id'], id=str(user['id']), platform_id=platform_id)
            player.user_id = str(user['id'])
            player.user_name = user.get('username')
            player.platform_id = platform_id
            player.rate_before = user.get('rate_before')
            player.rate_after = user.get('rate_after')
            if not player.rate_snapshot:
                player.rate_snapshot = user.get('rate_snapshot')

    def _guess_match_user(self, player, name):
        """Guess Voobly User from a player name."""
        try:
            user_id = str(self.platforms[PLATFORM_VOOBLY].find_user(name.lstrip('+')))
            get_unique(
                self.session,
                User, ['id', 'platform_id'],
                id=user_id,
                platform_id=PLATFORM_VOOBLY
            )
            player.user_id = user_id
        except RuntimeError as error:
            LOGGER.warning("failed to lookup Voobly user: %s", error)

    def _handle_series(self, series_id, series_name, map_data, log_id):
        """Handle series-related tasks."""
        if series_id and series_name:
            series = self.session.query(Series).get(series_id)
            if series:
                get_unique(
                    self.session,
                    SeriesMetadata,
                    ['series_id'],
                    name=series_name,
                    series_id=series.id
                )
                tournament = series.tournament
                event = tournament.event
                event_map = self.session.query(EventMap) \
                    .filter(EventMap.event_id == event.id) \
                    .filter(EventMap.name == map_data['name']) \
                    .one_or_none()
                if event_map:
                    event_map_id = event_map.id
                else:
                    event_map_id = None
                    LOGGER.warning("[m:%s] event map for %s not found: %s", log_id, event.id, map_data['name'])
                return series, tournament, event, event_map_id
        try:
            event_map = self.session.query(EventMap) \
                .filter(EventMap.name == map_data['name']) \
                .one()
            event_map_id = event_map.id
            LOGGER.info("[m:%s] guessed event %s for map %s", log_id, event_map.event_id, map_data['name'])
        except (NoResultFound, MultipleResultsFound):
            event_map_id = None
        return None, None, None, event_map_id

    def _handle_file(self, file_hash, data, version):
        """Handle file: compress and store."""
        compressed_filename = '{}{}'.format(file_hash, COMPRESSED_EXT)
        compressed_data = compress(io.BytesIO(data), version=version)
        destination = save_file(compressed_data, self.store_path, compressed_filename)
        LOGGER.info("[f:%s] copied to %s", file_hash[:LOG_ID_LENGTH], destination)
        return compressed_filename, len(compressed_data)



"""MGZ database schema."""

from sqlalchemy import (
    create_engine, Boolean, DateTime, Column,
    ForeignKey, Integer, Interval, String, Float, LargeBinary
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker, backref
from sqlalchemy.schema import ForeignKeyConstraint, UniqueConstraint

from mgzdb.util import get_utc_now


BASE = declarative_base()


def get_session(url):
    """Get SQL session."""
    engine = create_engine(url, echo=False)
    session = sessionmaker(bind=engine)()
    return session, engine


class Platform(BASE):
    """Multiplayer platform."""
    __tablename__ = 'platforms'
    id = Column(String, primary_key=True)
    name = Column(String)
    url = Column(String)
    match_url = Column(String)


class Person(BASE):
    """A known player."""
    __tablename__ = 'people'
    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    country = Column(String)


class Event(BASE):
    """Event (one or more tournaments)."""
    __tablename__ = 'events'
    id = Column(String, primary_key=True)
    name = Column(String)
    year = Column(Integer)


class Tournament(BASE):
    """A specific tournament (or stage of an event)."""
    __tablename__ = 'tournaments'
    id = Column(String, primary_key=True)
    name = Column(String)
    event_id = Column(String, ForeignKey('events.id'), index=True)
    event = relationship('Event', foreign_keys=[event_id], backref='tournaments')


class Round(BASE):
    """Tournament round."""
    __tablename__ = 'rounds'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    tournament_id = Column(String, ForeignKey('tournaments.id'), index=True)
    tournament = relationship('Tournament', foreign_keys=[tournament_id], backref='rounds')


class Series(BASE):
    """Series of matches."""
    __tablename__ = 'series'
    id = Column(String, primary_key=True)
    round_id = Column(Integer, ForeignKey('rounds.id'), index=True)
    round = relationship('Round', foreign_keys=[round_id], backref='series')
    played = Column(DateTime)
    tournament = relationship(
        'Tournament',
        secondary='rounds',
        primaryjoin='Series.round_id == Round.id',
        secondaryjoin='Tournament.id == Round.tournament_id',
        viewonly=True,
        uselist=False
    )


class Participant(BASE):
    """Series participants (team or single player)."""
    __tablename__ = 'participants'
    id = Column(Integer, primary_key=True)
    series_id = Column(String, ForeignKey('series.id'), index=True)
    series = relationship('Series', foreign_keys=[series_id], backref='participants')
    name = Column(String)
    score = Column(Integer)
    winner = Column(Boolean)


class Dataset(BASE):
    """AoC Dataset (a specific set of civs/balance/etc)."""
    __tablename__ = 'datasets'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class Civilization(BASE):
    """Civilization belonging to a dataset."""
    __tablename__ = 'civilizations'
    id = Column(Integer, primary_key=True)
    dataset_id = Column(Integer, ForeignKey('datasets.id'), primary_key=True)
    dataset = relationship('Dataset', foreign_keys=[dataset_id])
    name = Column(String, nullable=False)


class CivilizationBonus(BASE):
    """Bonus belonging to civilization."""
    __tablename__ = 'civilization_bonuses'
    id = Column(Integer, primary_key=True)
    civilization_id = Column(Integer)
    dataset_id = Column(Integer, ForeignKey('datasets.id'), index=True)
    type = Column(String)
    description = Column(String)
    civilization = relationship(
        'Civilization',
        foreign_keys=[civilization_id, dataset_id],
        backref='bonuses'
    )
    __table_args__ = (
        ForeignKeyConstraint(
            ['civilization_id', 'dataset_id'], ['civilizations.id', 'civilizations.dataset_id']
        ),
    )


class EventMap(BASE):
    """Event map."""
    __tablename__ = 'event_maps'
    id = Column(Integer, primary_key=True)
    event_id = Column(String, ForeignKey('events.id'), index=True)
    event = relationship('Event', foreign_keys=event_id, backref='maps')
    name = Column(String, index=True)
    zr = Column(Boolean)
    aoe2mapnet_id = Column(String)


class Map(BASE):
    """Builtin map."""
    __tablename__ = 'maps'
    id = Column(Integer, primary_key=True)
    dataset_id = Column(Integer, ForeignKey('datasets.id'), primary_key=True)
    dataset = relationship('Dataset', foreign_keys=[dataset_id])
    name = Column(String, index=True)


class Technology(BASE):
    """Technology."""
    __tablename__ = 'technologies'
    id = Column(Integer, primary_key=True)
    dataset_id = Column(Integer, ForeignKey('datasets.id'), primary_key=True)
    dataset = relationship('Dataset', foreign_keys=[dataset_id])
    name = Column(String)


class Object(BASE):
    """Object."""
    __tablename__ = 'objects'
    id = Column(Integer, primary_key=True)
    dataset_id = Column(Integer, ForeignKey('datasets.id'), primary_key=True)
    dataset = relationship('Dataset', foreign_keys=[dataset_id])
    name = Column(String)


class Terrain(BASE):
    """Terrain."""
    __tablename__ = 'terrain'
    id = Column(Integer, primary_key=True)
    dataset_id = Column(Integer, ForeignKey('datasets.id'), primary_key=True)
    name = Column(String)
    color_level = Column(String(7))
    color_up = Column(String(7))
    color_down = Column(String(7))


class GameType(BASE):
    """Game type."""
    __tablename__ = 'game_types'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class StartingResources(BASE):
    """Starting resoruces."""
    __tablename__ = 'starting_resources'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class VictoryCondition(BASE):
    """Victory conditions."""
    __tablename__ = 'victory_conditions'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class MapRevealChoice(BASE):
    """Map reveal choices."""
    __tablename__ = 'map_reveal_choices'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class StartingAge(BASE):
    """Starting Ages."""
    __tablename__ = 'starting_ages'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class Difficulty(BASE):
    """Difficulties."""
    __tablename__ = 'difficulties'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class Speed(BASE):
    """Speeds."""
    __tablename__ = 'speeds'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class MapSize(BASE):
    """Map sizes."""
    __tablename__ = 'map_sizes'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class PlayerColor(BASE):
    """Player color."""
    __tablename__ = 'player_colors'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    hex = Column(String)


class File(BASE):
    """Represent File."""
    __tablename__ = 'files'
    id = Column(Integer, primary_key=True)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    match = relationship('Match', foreign_keys=[match_id])
    hash = Column(String, unique=True, nullable=False)
    filename = Column(String, nullable=False)
    original_filename = Column(String)
    modified = Column(DateTime)
    encoding = Column(String)
    language = Column(String, index=True)
    size = Column(Integer, nullable=False)
    compressed_size = Column(Integer, nullable=False)
    owner_number = Column(Integer, nullable=False)
    owner = relationship('Player', foreign_keys=[match_id, owner_number], viewonly=True, post_update=True)
    reference = Column(String)
    added = Column(DateTime, default=get_utc_now)
    parser_version = Column(String, nullable=False)
    __table_args__ = (
        ForeignKeyConstraint(['match_id', 'owner_number'], ['players.match_id', 'players.number'], ondelete='cascade'),
    )


class Match(BASE):
    """Represents Match."""
    __tablename__ = 'matches'
    id = Column(Integer, primary_key=True)
    hash = Column(String, unique=True)
    series_id = Column(String, ForeignKey('series.id'), index=True)
    tournament_id = Column(String, ForeignKey('tournaments.id'), index=True)
    event_id = Column(String, ForeignKey('events.id'), index=True)
    series = relationship('Series', foreign_keys=series_id, backref='matches')
    tournament = relationship('Tournament', foreign_keys=tournament_id, backref='matches')
    event = relationship('Event', foreign_keys=event_id, backref='matches')
    players = relationship('Player', back_populates='match', cascade='all, delete')
    teams = relationship('Team', foreign_keys='Team.match_id', cascade='all, delete')
    winning_team_id = Column(Integer)
    winning_team = relationship('Player', primaryjoin='and_(Player.match_id==Match.id, ' \
                                                      'Player.team_id==Match.winning_team_id)')
    losers = relationship('Player', primaryjoin='and_(Player.match_id==Match.id, ' \
                                                'Player.team_id!=Match.winning_team_id)')
    files = relationship('File', foreign_keys='File.match_id', cascade='all, delete, delete-orphan', post_update=True)
    version = Column(String)
    minor_version = Column(String)
    version_id = Column(Integer, ForeignKey('versions.id'))
    game_version = Column(String)
    save_version = Column(Float)
    build = Column(String)
    dataset_id = Column(Integer, ForeignKey('datasets.id'), index=True)
    dataset_version = Column(String)
    dataset = relationship('Dataset', foreign_keys=dataset_id)
    platform_id = Column(String, ForeignKey('platforms.id'), index=True)
    platform = relationship('Platform', foreign_keys=platform_id)
    platform_metadata = Column(String)
    ladder_id = Column(Integer)
    ladder = relationship('Ladder', foreign_keys=[ladder_id, platform_id], viewonly=True)
    rated = Column(Boolean)
    lobby_name = Column(String)
    lobby_private = Column(Boolean)
    lobby_opened = Column(DateTime)
    server = Column(String)
    builtin_map_id = Column(Integer)
    builtin_map = relationship('Map', foreign_keys=[builtin_map_id, dataset_id], backref='matches', viewonly=True)
    map_size_id = Column(Integer, ForeignKey('map_sizes.id'), index=True)
    map_size = relationship('MapSize', foreign_keys=map_size_id)
    map_name = Column(String, index=True)
    map_tiles = Column(LargeBinary)
    event_map_id = Column(Integer, ForeignKey('event_maps.id'), index=True)
    event_map = relationship('EventMap', foreign_keys=event_map_id, backref='matches')
    rms_zr = Column(Boolean)
    rms_custom = Column(Boolean)
    rms_seed = Column(Integer)
    guard_state = Column(Boolean)
    fixed_positions = Column(Boolean)
    direct_placement = Column(Boolean)
    effect_quantity = Column(Boolean)
    water_percent = Column(Float)
    played = Column(DateTime, index=True)
    added = Column(DateTime, default=get_utc_now)
    platform_match_id = Column(String, unique=True)
    duration = Column(Interval)
    completed = Column(Boolean)
    restored = Column(Boolean)
    postgame = Column(Boolean)
    type_id = Column(Integer, ForeignKey('game_types.id'), index=True)
    type = relationship('GameType', foreign_keys=type_id)
    difficulty_id = Column(Integer, ForeignKey('difficulties.id'), index=True)
    difficulty = relationship('Difficulty', foreign_keys=difficulty_id)
    population_limit = Column(Integer)
    map_reveal_choice_id = Column(Integer, ForeignKey('map_reveal_choices.id'), index=True)
    map_reveal_choice = relationship('MapRevealChoice', foreign_keys=map_reveal_choice_id)
    cheats = Column(Boolean)
    speed_id = Column(Integer, ForeignKey('speeds.id'), index=True)
    speed = relationship('Speed', foreign_keys=speed_id)
    lock_teams = Column(Boolean)
    mirror = Column(Boolean)
    diplomacy_type = Column(String, nullable=False, index=True)
    team_size = Column(String, nullable=False, index=True)
    starting_resources_id = Column(Integer, ForeignKey('starting_resources.id'), index=True)
    starting_resources = relationship('StartingResources', foreign_keys=starting_resources_id)
    starting_age_id = Column(Integer, ForeignKey('starting_ages.id'), index=True)
    starting_age = relationship('StartingAge', foreign_keys=starting_age_id)
    starting_town_centers = Column(Integer)
    starting_palisades = Column(Boolean)
    starting_walls = Column(Boolean)
    victory_condition_id = Column(Integer, ForeignKey('victory_conditions.id'), index=True)
    victory_condition = relationship('VictoryCondition', foreign_keys=victory_condition_id)
    treaty_length = Column(Integer)
    team_together = Column(Boolean)
    all_technologies = Column(Boolean)
    lock_speed = Column(Boolean)
    multiqueue = Column(Boolean)
    has_playback = Column(Boolean)
    object_instances = relationship('ObjectInstance', foreign_keys='ObjectInstance.match_id', cascade='all, delete, delete-orphan')
    research = relationship('Research', foreign_keys='Research.match_id', cascade='all, delete, delete-orphan')
    timeseries = relationship('Timeseries', foreign_keys='Timeseries.match_id', cascade='all, delete, delete-orphan')
    market = relationship('Market', foreign_keys='Market.match_id', cascade='all, delete, delete-orphan')
    state_reader_version = Column(String)
    state_reader_interval = Column(Integer)
    state_reader_runtime = Column(Interval)
    hc = Column(Boolean)
    __table_args__ = (
        ForeignKeyConstraint(['ladder_id', 'platform_id'], ['ladders.id', 'ladders.platform_id']),
        ForeignKeyConstraint(['builtin_map_id', 'dataset_id'], ['maps.id', 'maps.dataset_id'])
    )


class Team(BASE):
    """Represent a team."""
    __tablename__ = 'teams'
    team_id = Column(Integer, primary_key=True)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), primary_key=True)
    winner = Column(Boolean)
    match = relationship('Match', foreign_keys=match_id)


class Player(BASE):
    """Represent Player in a Match."""
    __tablename__ = 'players'
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), primary_key=True, index=True)
    name = Column(String, nullable=False)
    number = Column(Integer, nullable=False, primary_key=True, index=True)
    color_id = Column(Integer, ForeignKey('player_colors.id'), nullable=False, index=True)
    color = relationship('PlayerColor', foreign_keys=color_id)
    platform_id = Column(String, ForeignKey('platforms.id'), index=True)
    platform = relationship('Platform', foreign_keys=platform_id)
    user_id = Column(String, index=True)
    user_name = Column(String, index=True)
    user = relationship('User', foreign_keys=[user_id, platform_id], viewonly=True)
    match = relationship('Match', foreign_keys=[match_id], viewonly=True)
    team_id = Column(Integer)
    team = relationship('Team', foreign_keys=[match_id, team_id], backref='members', viewonly=True, single_parent=True, post_update=True)
    dataset_id = Column(Integer, ForeignKey('datasets.id'))
    dataset = relationship('Dataset', foreign_keys=[dataset_id])
    civilization_id = Column(Integer, index=True)
    civilization = relationship('Civilization', foreign_keys=[dataset_id, civilization_id], backref='players', viewonly=True)
    start_x = Column(Integer)
    start_y = Column(Integer)
    human = Column(Boolean)
    winner = Column(Boolean)
    mvp = Column(Boolean)
    cheater = Column(Boolean)
    score = Column(Integer)
    rate_before = Column(Float)
    rate_after = Column(Float, index=True)
    rate_snapshot = Column(Float, index=True)
    military_score = Column(Integer)
    units_killed = Column(Integer)
    hit_points_killed = Column(Integer)
    units_lost = Column(Integer)
    buildings_razed = Column(Integer)
    hit_points_razed = Column(Integer)
    buildings_lost = Column(Integer)
    units_converted = Column(Integer)
    economy_score = Column(Integer)
    food_collected = Column(Integer)
    wood_collected = Column(Integer)
    stone_collected = Column(Integer)
    gold_collected = Column(Integer)
    tribute_sent = Column(Integer)
    tribute_received = Column(Integer)
    trade_gold = Column(Integer)
    relic_gold = Column(Integer)
    technology_score = Column(Integer)
    feudal_time = Column(Interval)
    castle_time = Column(Interval)
    imperial_time = Column(Interval)
    explored_percent = Column(Integer)
    research_count = Column(Integer)
    research_percent = Column(Integer)
    society_score = Column(Integer)
    total_wonders = Column(Integer)
    total_castles = Column(Integer)
    total_relics = Column(Integer)
    villager_high = Column(Integer)
    __table_args__ = (
        ForeignKeyConstraint(['match_id', 'team_id'], ['teams.match_id', 'teams.team_id'], ondelete='cascade'),
        ForeignKeyConstraint(['civilization_id', 'dataset_id'], ['civilizations.id', 'civilizations.dataset_id']),
        ForeignKeyConstraint(['user_id', 'platform_id'], ['users.id', 'users.platform_id']),
    )


class User(BASE):
    """Represents a Platform User."""
    __tablename__ = 'users'
    id = Column(String, primary_key=True)
    platform_id = Column(String, ForeignKey('platforms.id'), primary_key=True)
    person_id = Column(Integer, ForeignKey('people.id'), index=True)
    added = Column(DateTime, default=get_utc_now)
    __table_args__ = (
        UniqueConstraint('id', 'platform_id'),
    )


class Ladder(BASE):
    """Represents a platform Ladder."""
    __tablename__ = 'ladders'
    id = Column(Integer, primary_key=True)
    platform_id = Column(String, ForeignKey('platforms.id'), primary_key=True)
    platform = relationship('Platform', foreign_keys=[platform_id], backref='ladders')
    name = Column(String, nullable=False)


class SeriesMetadata(BASE):
    """Represents series metadata."""
    __tablename__ = 'series_metadata'
    id = Column(Integer, primary_key=True)
    series_id = Column(String, ForeignKey('series.id'), index=True)
    series = relationship('Series', foreign_keys=[series_id], backref=backref('metadata', uselist=False))
    name = Column(String)


class Chat(BASE):
    """Represents chat between players."""
    __tablename__ = 'chat'
    id = Column(Integer, primary_key=True)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    player_number = Column(Integer)
    timestamp = Column(Interval)
    player = relationship('Player', foreign_keys=[match_id, player_number], viewonly=True)
    match = relationship('Match', foreign_keys=[match_id])
    message = Column(String)
    origination = Column(String)
    audience = Column(String)
    __table_args__ = (
        ForeignKeyConstraint(['match_id', 'player_number'], ['players.match_id', 'players.number'], ondelete='cascade'),
    )


class Timeseries(BASE):
    """Represents timeseries data."""
    __tablename__ = 'timeseries'
    id = Column(Integer, primary_key=True)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    player_number = Column(Integer)
    timestamp = Column(Interval)
    player = relationship('Player', foreign_keys=[match_id, player_number], viewonly=True, backref='chat')
    match = relationship('Match', foreign_keys=[match_id], backref='chat')
    population = Column(Float)
    military = Column(Float)
    percent_explored = Column(Float)
    headroom = Column(Integer)
    food = Column(Integer)
    wood = Column(Integer)
    stone = Column(Integer)
    gold = Column(Integer)
    total_housed_time = Column(Interval)
    total_popcapped_time = Column(Interval)
    relics_captured = Column(Integer)
    relic_gold = Column(Integer)
    trade_profit = Column(Integer)
    tribute_sent = Column(Integer)
    tribute_received = Column(Integer)
    total_food = Column(Integer)
    total_wood = Column(Integer)
    total_stone = Column(Integer)
    total_gold = Column(Integer)
    value_spent_objects = Column(Integer)
    value_spent_research = Column(Integer)
    value_lost_units = Column(Integer)
    value_lost_buildings = Column(Integer)
    value_current_units = Column(Integer)
    value_current_buildings = Column(Integer)
    value_objects_destroyed = Column(Integer)
    kills = Column(Integer)
    razes = Column(Integer)
    deaths = Column(Integer)
    __table_args__ = (
        ForeignKeyConstraint(['match_id', 'player_number'], ['players.match_id', 'players.number'], ondelete='cascade'),
    )


class Research(BASE):
    """Research."""
    __tablename__ = 'research'
    id = Column(Integer, primary_key=True)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    dataset_id = Column(Integer, ForeignKey('datasets.id'), index=True)
    dataset = relationship('Dataset', foreign_keys=dataset_id)
    player_number = Column(Integer)
    player = relationship('Player', foreign_keys=[match_id, player_number], viewonly=True, backref='researches')
    match = relationship('Match', foreign_keys=[match_id], backref='researches')
    technology_id = Column(Integer)
    technology = relationship('Technology', foreign_keys=[technology_id, dataset_id], viewonly=True)
    started = Column(Interval)
    finished = Column(Interval)
    __table_args__ = (
        ForeignKeyConstraint(['match_id', 'player_number'], ['players.match_id', 'players.number'], ondelete='cascade'),
        ForeignKeyConstraint(['dataset_id', 'technology_id'], ['technologies.dataset_id', 'technologies.id'])
    )


class ObjectInstance(BASE):
    """Instance of object."""
    __tablename__ = 'object_instances'
    id = Column(Integer, primary_key=True)
    instance_id = Column(Integer, index=True)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    dataset_id = Column(Integer, ForeignKey('datasets.id'), index=True)
    dataset = relationship('Dataset', foreign_keys=dataset_id)
    initial_player_number = Column(Integer, index=True)
    initial_object_id = Column(Integer, index=True)
    initial_class_id = Column(Integer, index=True)
    states = relationship('ObjectInstanceState', foreign_keys='ObjectInstanceState.match_id', cascade='all, delete, delete-orphan')
    created = Column(Interval)
    destroyed = Column(Interval)
    destroyed_by_instance_id = Column(Integer, index=True)
    deleted = Column(Boolean)
    created_x = Column(Float)
    created_y = Column(Float)
    destroyed_x = Column(Float)
    destroyed_y = Column(Float)
    destroyed_building_percent = Column(Float)
    building_started = Column(Interval)
    building_completed = Column(Interval)
    total_idle_time = Column(Interval)
    __table_args__ = (
        UniqueConstraint('match_id', 'instance_id'),
    )


class ObjectInstanceState(BASE):
    """Instance state."""
    __tablename__ = 'object_instance_states'
    id = Column(Integer, primary_key=True)
    timestamp = Column(Interval)
    instance_id = Column(Integer, index=True)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    dataset_id = Column(Integer, ForeignKey('datasets.id'), index=True)
    dataset = relationship('Dataset', foreign_keys=dataset_id)
    player_number = Column(Integer)
    player = relationship('Player', foreign_keys=[match_id, player_number], viewonly=True, backref='objects')
    match = relationship('Match', foreign_keys=[match_id], backref='object_states')
    object_id = Column(Integer, index=True)
    object = relationship('Object', foreign_keys=[object_id, dataset_id], viewonly=True)
    class_id = Column(Integer, index=True)
    researching_technology_id = Column(Integer)
    __table_args__ = (
        ForeignKeyConstraint(['match_id', 'player_number'], ['players.match_id', 'players.number'], ondelete='cascade'),
        ForeignKeyConstraint(['dataset_id', 'object_id'], ['objects.dataset_id', 'objects.id']),
        ForeignKeyConstraint(['instance_id', 'match_id'], ['object_instances.instance_id', 'object_instances.match_id'], ondelete='cascade'),
        ForeignKeyConstraint(['dataset_id', 'researching_technology_id'], ['technologies.dataset_id', 'technologies.id'])
    )


class Market(BASE):
    """Market coefficients."""
    __tablename__ = 'market'
    id = Column(Integer, primary_key=True)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    match = relationship('Match', foreign_keys=[match_id])
    timestamp = Column(Interval)
    wood = Column(Float)
    stone = Column(Float)
    food = Column(Float)


class ActionLog(BASE):
    """Action log."""
    __tablename__ = 'action_log'
    id = Column(Integer, primary_key=True)
    timestamp = Column(Interval, index=True)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    match = relationship('Match', foreign_keys=[match_id])
    player_number = Column(Integer, index=True)
    action_id = Column(Integer, ForeignKey('actions.id'), index=True)
    action = relationship('Action', foreign_keys=[action_id], viewonly=True)
    action_x = Column(Float)
    action_y = Column(Float)


class Action(BASE):
    """Actions."""
    __tablename__ = 'actions'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class Tribute(BASE):
    """Tribute."""
    __tablename__ = 'tribute'
    id = Column(Integer, primary_key=True)
    timestamp = Column(Interval)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    player_number = Column(Integer, index=True)
    target_player_number = Column(Integer)
    resource_id = Column(Integer, ForeignKey('resources.id'))
    amount = Column(Float)
    fee = Column(Float)


class Transaction(BASE):
    """Market commodity transactions."""
    __tablename__ = 'transactions'
    id = Column(Integer, primary_key=True)
    timestamp = Column(Interval)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    player_number = Column(Integer, index=True)
    action_id = Column(Integer, ForeignKey('actions.id'))
    resource_id = Column(Integer, ForeignKey('resources.id'))
    amount = Column(Integer)


class Formation(BASE):
    """Formation usage."""
    __tablename__ = 'formations'
    id = Column(Integer, primary_key=True)
    timestamp = Column(Interval)
    match_id = Column(Integer, ForeignKey('matches.id', ondelete='cascade'), index=True)
    player_number = Column(Integer, index=True)
    formation_id = Column(Integer, ForeignKey('formation_types.id'))


class FormationType(BASE):
    """Formation types."""
    __tablename__ = 'formation_types'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class Resource(BASE):
    """Resource."""
    __tablename__ = 'resources'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class Version(BASE):
    """Version."""
    __tablename__ = 'versions'
    id = Column(Integer, primary_key=True)
    name = Column(String)


class HiddenCupPlayers(BASE):
    """HC player."""
    __tablename__ = 'hc'
    hc_name = Column(String, primary_key=True)
    person_name = Column(String, primary_key=True)
    event_id = Column(String)



"""Record extracted data."""
import logging
from datetime import timedelta

from mgz import fast
from mgzdb.schema import (
    Timeseries, Research, ObjectInstance, Market,
    ObjectInstanceState, ActionLog, Tribute, Transaction,
    Formation
)


LOGGER = logging.getLogger(__name__)
ALLOWED = False
ALLOWED_LADDERS = [131] #, 132]
ALLOWED_RATE = 0
SUPPORTED_DATASETS = [1]
STARTING_SUPPORTED_DATASETS = [0, 1, 100]
SAMPLE_INTERVAL = 30000


def allow_extraction(players, ladder_id, dataset_id):
    """Check whether extraction should be attempted."""
    rate_sum = sum([p.get('rate_snapshot', 0) for p in players if p.get('rate_snapshot')])
    rate_avg = rate_sum / len(players)
    return (
        ALLOWED and
        ladder_id in ALLOWED_LADDERS and
        dataset_id in SUPPORTED_DATASETS and
        rate_avg >= ALLOWED_RATE
    )


def save_starting_state(session, objects, match_id, dataset_id):
    """Save starting state."""
    if dataset_id not in STARTING_SUPPORTED_DATASETS:
        return False
    objs = []
    for record in objects:
        objs.append(ObjectInstance(
            match_id=match_id,
            dataset_id=dataset_id,
            instance_id=record['instance_id'],
            initial_object_id=record['object_id'],
            initial_class_id=record['class_id'],
            initial_player_number=record['player_number'],
            created=timedelta(milliseconds=0),
            created_x=record['x'],
            created_y=record['y']
        ))
    session.bulk_save_objects(objs)
    session.commit()
    return True


def save_extraction(session, summary, ladder_id, match_id, dataset_id, log_id, force=False):
    """Commit extraction data when available."""
    if not force and (not summary.can_playback() or not allow_extraction(list(summary.get_players()), ladder_id, dataset_id)):
        LOGGER.info("[m:%s] skipping full extraction, did not meet requirements", log_id)
        return False, {}
    LOGGER.info("[m:%s] starting full extraction", log_id)
    try:
        extracted = summary.extract(SAMPLE_INTERVAL)
    except RuntimeError as error:
        LOGGER.warning("[m:%s] failed to complete extraction: %s", log_id, error)
        return False, {}
    objs = []
    for record in extracted['timeseries']:
        record['timestamp'] = timedelta(milliseconds=record['timestamp'])
        record['total_housed_time'] = timedelta(milliseconds=record['total_housed_time'])
        record['total_popcapped_time'] = timedelta(milliseconds=record['total_popcapped_time'])
        objs.append(Timeseries(match_id=match_id, **record))
    for record in extracted['market']:
        record['timestamp'] = timedelta(milliseconds=record['timestamp'])
        objs.append(Market(match_id=match_id, **record))
    for record in extracted['research']:
        record['started'] = timedelta(milliseconds=record['started'])
        record['finished'] = timedelta(milliseconds=record['finished']) if record['finished'] else None
        objs.append(Research(match_id=match_id, dataset_id=dataset_id, **record))
    for record in extracted['objects']:
        record['created'] = timedelta(milliseconds=record['created'])
        record['destroyed'] = timedelta(milliseconds=record['destroyed']) if record['destroyed'] else None
        record['building_started'] = timedelta(milliseconds=record['building_started']) if record['building_started'] else None
        record['building_completed'] = timedelta(milliseconds=record['building_completed']) if record['building_completed'] else None
        record['total_idle_time'] = timedelta(milliseconds=record['total_idle_time']) if record['total_idle_time'] else None
        objs.append(ObjectInstance(match_id=match_id, dataset_id=dataset_id, **record))
    for record in extracted['state']:
        record['timestamp'] = timedelta(milliseconds=record['timestamp'])
        objs.append(ObjectInstanceState(match_id=match_id, dataset_id=dataset_id, **record))
    for (timestamp, action_type, payload) in extracted['actions']:
        if action_type == fast.Action.TRIBUTE:
            objs.append(Tribute(
                match_id=match_id,
                timestamp=timedelta(milliseconds=timestamp),
                player_number=payload['player_id'],
                target_player_number=payload['player_id_to'],
                resource_id=payload['resource_id'],
                amount=payload['amount'],
                fee=payload['fee']
            ))
        if action_type in [fast.Action.BUY, fast.Action.SELL]:
            objs.append(Transaction(
                match_id=match_id,
                timestamp=timedelta(milliseconds=timestamp),
                action_id=action_type.value,
                player_number=payload['player_id'],
                resource_id=payload['resource_id'],
                amount=payload['amount']
            ))
        if summary.get_diplomacy().get('type') == '1v1':
            objs.append(ActionLog(
                match_id=match_id,
                timestamp=timedelta(milliseconds=timestamp),
                action_id=action_type.value,
                player_number=payload.get('player_id'),
                action_x=payload.get('x'),
                action_y=payload.get('y')
            ))
            if action_type == fast.Action.FORMATION:
                objs.append(Formation(
                    match_id=match_id,
                    timestamp=timedelta(milliseconds=timestamp),
                    formation_id=payload.get('formation_id'),
                    player_number=payload.get('player_id')
                ))

    session.bulk_save_objects(objs)
    session.commit()
    LOGGER.info("[m:%s] completed full extraction", log_id)
    return True, dict(version=extracted['version'], interval=SAMPLE_INTERVAL, runtime=extracted['runtime'])
